#!/usr/bin/ruby
# $Id$
# -*- ruby -*- vim:set ft=ruby et sw=2 sts=2:

require 'io/mixins'
require 'stringio'
require 'forwardable'
require 'test/unit'

class SimpleIO
  include IO::Readable
  include IO::Writable
  include IO::Seekable
  extend  IO::Openable
  extend  Forwardable
  def initialize(*args)
    @string = StringIO.new(*args)
  end
  def sysread(length)
    @string.read(length) or raise EOFError, "end of file reached", caller
  end
  def syswrite(chunk)
    @string.write(chunk)
  end
  def sysseek(offset, whence = IO::SEEK_SET)
    @string.seek(offset,whence)
    @string.tell
  end
  DELEGATORS = [:close_read,:close_write,:closed_read?,:closed_write?,:reopen]
  def_delegators :@string, *DELEGATORS
  # def close_read()   @string.close_read    end
  # def close_write()  @string.close_write   end
  # def closed_read?() @string.closed_read?  end
  # def closed_write?()@string.closed_write? end
  # def close()        @string.close         end
end

require File.join(File.dirname(__FILE__),'rot13_filter')
require File.join(File.dirname(__FILE__),'io_loop')
require File.join(File.dirname(__FILE__),'read_joiner')

class PairTester
  def initialize(test,initial,*args)
    @test = test
    @string1 = initial.dup
    @string2 = initial.dup
    @s1 = StringIO.new(@string1,*args)
    @s2 = SimpleIO.new(@string2,*args)
  end

  def perform(*args)
    @s1.__send__(*args)
    @s2.__send__(*args)
  end

  def assert_equal
    @test.assert_equal @string1, @string2
  end

  def method_missing(*args)
    value1 = @s1.__send__(*args) rescue $!.class
    value2 = @s2.__send__(*args) rescue $!.class
    value2 = value1 if value1 == @s1 && value2 == @s2
    @test.assert_equal value1, value2
    value1
  end

end

class IOMixinsTest < Test::Unit::TestCase

  def test_minimal
    SimpleIO.open(":)\n") do |io|
      assert_nothing_raised { io.gets }
      assert_raise(NotImplementedError) { io.fcntl }
    end
  end

  def test_enough_methods
    junk = Object.instance_methods | Enumerable.instance_methods
    good = IO.instance_methods(false) & StringIO.instance_methods(false)
    mine = SimpleIO.instance_methods
    assert_equal [], good - junk - mine
  end

  def test_io_loop
    string = "foo\nbar\n"
    io = IOLoop.open(string)
    assert_equal ?f, io.getc
    io.ungetc ?g
    assert_equal ?g, io.getc
    io.putc "lqqqqq" # Should use first character
    assert_equal "o\n", io.gets
    io.puts "baz\n"
    assert_equal "flo\n", io.readline
    assert_equal "baz\n", io.readline
    io.close
    io.instance_variable_set("@closed_read",false)
    assert_equal true, io.closed_read?
    assert_raise(IOError) { io.read }
    io = IOLoop.open("foo")
    assert_raise(IOError) { io.read }
  end

  def empty_file
    File.open(RUBY_PLATFORM =~ /win32/ ? "NUL:" : "/dev/null")
  end

  def empty_string
    StringIO.open("")
  end

  def empty_simple
    SimpleIO.new("")
  end

  def assert_identical_response(object1,object2,method,*args,&block)
    value1 = object1.send(method,*args,&block) rescue $!.class
    value2 = object2.send(method,*args,&block) rescue $!.class
    assert_equal value1, value2, "#{object1.class} gave #{value1.inspect} while #{object2.class} gave #{value2.inspect}"
  end

  def initialize_both(*args)
    @both = PairTester.new(self,*args)
  end

  def test_simple_io_empty_read
    [lambda {empty_file},lambda {empty_string}].each do |l|
      assert_identical_response l.call,empty_simple,:read
      assert_identical_response l.call,empty_simple,:read,0
      assert_identical_response l.call,empty_simple,:read,2
    end
  end

  def test_simple_io_read
    initialize_both("Hello\nworld!")
    @both.eof
    @both.tty?
    @both.lineno
    @both.read
    @both.eof?
    @both.seek(2)
    @both.pos
    @both.rewind
    @both.tell
    @both.gets
    @both.lineno
    @both.read(4, "buffer")
    @both.tell
    @both.read(0)
    @both.read(6)
    @both.read
    @both.eof
    @both.read(0)
    @both.rewind
    @both.lineno
    @both.read(8)
    @both.read(0)
    @both.read(1)
    @both.readchar
    @both.close_read
    @both.close
    @both.close_write
    @both.closed_read?
    @both.closed_write?
    @both.sync
    @both.closed?
  end

  def test_simple_io_gets
    # io = SimpleIO.open(StringIO.new(<<-EOF))
    string = <<-EOF


Hello world!


Chunky
bacon!



Goodbye world!
    EOF
    string.chomp!
    initialize_both(string)
    @both.gets ""
    @both.readline ""
    @both.map {|x|x}
    @both.close
  end

  def test_simple_io_write
    initialize_both("x","w")
    @both.eof
    @both.putc(42)
    @both.eof # => IOError
    @both.print("Hello ")
    @both.binmode
    @both.puts("world")
    @both << ":-"
    @both << ?)
    @both.fsync
    @both.printf("%4d",17)
    @both.putc(1..3) # => TypeError
    @both.putc(:bad) # => TypeError
    @both.pos
    @both.flush
    @both.pos = 6
    @both.putc(?_)
    @both.close
    @both.assert_equal
  end

  def test_simple_io_seek
    initialize_both("Look ma!","r+")
    @both.pos
    @both.getc
    @both.tell
    @both.seek(3,IO::SEEK_SET)
    @both.read(2)
    @both.ungetc(?z)
    @both.tell
    @both.seek(-2,IO::SEEK_CUR)
    @both.getc
    @both.pos
    @both.seek(-3,IO::SEEK_END)
    @both.tell
    @both.getc
  end

  def test_missing_methods
    readable = Class.new do
      include IO::Readable
    end
    writable = Class.new do
      include IO::Writable
      def close_write
        if @closed_write
          raise IOError
        else
          @closed_write = true
        end
      end
      def closed_write?
        !!@closed_write
      end
    end
    streamable = Class.new(writable) do
      include IO::Readable
      def close_read
        if @closed_read
          raise IOError
        else
          @closed_read = true
        end
      end
      def closed_read?
        !!@closed_read
      end
    end
    assert_raise(NotImplementedError) {readable.new.close}
    assert_raise(NotImplementedError) {readable.new.closed?}
    w = writable.new
    assert_equal false, w.closed_write?
    assert_nothing_thrown {w.close}
    assert_raise(IOError) {w.close}
    assert_equal true, w.closed?
    s = streamable.new
    assert_equal false, s.closed?
    assert_nothing_thrown {s.close_read}
    assert_nothing_thrown {s.close}
    assert_raise(IOError) {s.close_write}
    assert_raise(IOError) {s.close}
    assert_equal true, s.closed?
  end

  def test_open_readers
    assert_raise(EOFError) { SimpleIO.open("") {|io|io.sysread(1)}}
    assert_equal "READ", SimpleIO.read("README",4)
    assert_equal "x" * 1025, SimpleIO.read("x"*1025)
    assert_equal ["foo\n","bar"], SimpleIO.readlines("foo\nbar")
    assert_raise(LocalJumpError) { SimpleIO.foreach("foo\nbar") }
  end

  def test_rot13
    string = "Puhaxl onpba!"
    stringio = StringIO.open(string,"r")
    filtered = Rot13Filter.open(stringio)
    assert_equal "Chunky bacon!", filtered.read
    stringio = StringIO.open(string,"r+")
    filtered = Rot13Filter.open(stringio)
    filtered.write("Rotten")
    assert_equal "Ebggra onpba!", string
  end

  def test_joiner
    joiner = ReadJoiner.new(StringIO.new("fo"),StringIO.new("o\nbar\n"))
    assert_equal "foo\n", joiner.readline
    assert_nil   joiner.close
    assert_equal true, joiner.closed?
  end

end
